#! /usr/bin/env python
#
#    Implements a simple face-vertex representation of a model
#    Copyright (C) 2002  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import model
import math
import edgebreaker

try:
    import pygame
    from pygame.locals import *
    from OpenGL.GL import *
except:
    print 'fvmodel requires pygame and pyopengl'
    raise SystemExit


def _cross(a, b):
    return (a[1]*b[2]-a[2]*b[1], 
            a[2]*b[0]-a[0]*b[2], 
            a[0]*b[1]-a[1]*b[0])


class FVlist(model.Model):
    """Handles loading and drawing of a simple model format"""

    def __init__(self, filename=None):
        """SMF([filename]) optionally loads model stored in passed filename"""

        model.Model.__init__(self)

        self.vertices = []          # list of vertices
        self.colors = []            # corresponding colors
        self.faces = []             # list of references to vertices
        self.dual_face_cycles = []  # list of references to faces, not tris
        self.normals = []           # normals correspond to each face
        self.texture = None
        self.texturecoords = []     # S,T corresponds to each vertex

        if filename:
            self.load(filename)

    def load(self, filename):
        """load(self, filename) -> None

        Load a model in simplified OBJ format"""

        if filename.endswith('.gz'):
            try:
                import gzip
                f = gzip.open(filename)
            except ImportError:
                raise SystemExit("Viewing gzipped files requires module gzip")
        else:
            f = open(filename)
        linecount = 0
        for line in f.readlines():
            linecount += 1
            items = line.split()
            # v X Y Z defines a vertex at (x, y, z)
            if len(items) == 4 and items[0] == 'v':
                self.vertices.append(map(lambda x:float(x), items[1:4]))
            # f A B C defines a face using vertices A, B, C
            elif len(items) == 4 and items[0] == 'f':
                self.faces.append(map(lambda x:int(x)-1, items[1:4]))
            # c R G B defines a color for corresponding vertex
            elif len(items) == 4 and items[0] == 'c':
                self.colors.append(map(lambda x:float(x), items[1:4]))
            # t S T defines a texture coordinate for corresponding vertex
            elif len(items) == 3 and items[0] == 't':
                self.texturecoords.append(map(lambda x:float(x), items[1:3]))
            # t filename defines a texture for the model
            #            should be 2**k x 2**k pixels
            elif len(items) == 2 and items[0] == 't':
                if self.texture:
                    raise RuntimeError("Can't handle multiple textures")
                self.texture = items[1]
            elif line[0] == '#' or len(items) == 0:
                pass
            else:
                raise RuntimeError("Invalid syntax on line %d '%s'"%(linecount, line))
        f.close()

        self.calculate_normals()

        if self.texture:
            if not GLU:
                raise NotImplementedError("textures require mipmaps require OpenGL.GLU")

            # load and prepare texture image for opengl
            img = pygame.image.load(self.texture)
            w, h = img.get_width(), img.get_height()
            rgb = pygame.image.tostring(img, "RGB", 0)

            #assign a texture
            self.textureid = glGenTextures(1)
            glBindTexture(GL_TEXTURE_2D, self.textureid)
            #glPixelStorei(GL_UNPACK_ALIGNMENT,1)


            #build MIPMAP levels
            gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, w, h, GL_RGB, GL_UNSIGNED_BYTE, rgb)
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)

            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)

    def calculate_normals(self):
        """calculate_normals(self) -> None

        Calculates the normals from a proper fvlist"""

        self.normals = []
        for A,B,C in self.faces:
            A, B, C = self.vertices[A], self.vertices[B], self.vertices[C]
            AB = A[0]-B[0], A[1]-B[1], A[2]-B[2]
            AC = A[0]-C[0], A[1]-C[1], A[2]-C[2]
            n = _cross(AB, AC)
            mag = math.sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2])
            try:
                n = n[0]/mag, n[1]/mag, n[2]/mag
            except ZeroDivisionError:
                n = (0.0, 0.0, 0.0)
            self.normals.append(n)


    def build_ovtable(self):
        """build_ovtable(self) -> None

        Build the OVTable self.ovtable of the model"""

        self.ovtable = [ [] for i in range(len(self.faces)) ]

        sf = self.faces
        vtof = self.build_vtof()
        self.build_dual_graph()

        for face in range(len(self.faces)):
            for v in sf[face]:

                lv, rv = [vx for vx in sf[face] if vx != v]
                llist = self.dual_face_cycles[lv]
                rlist = self.dual_face_cycles[rv]
                l = [l for ls in llist for l in ls]
                r = [r for rs in rlist for r in rs]
                ofaces = [f for f in l if f in r and f != face]

                ov = [o for o in sf[ofaces[0]] if o not in (lv, rv)][0]
                self.ovtable[face].append(ov)

    def save_ovtable(self, filename):
        """save_ovtable(self, filename) -> None

        Saves the OVTable to the file specified.  If the filename ends in .gz,
        the output will be gzip compressed."""

        # first open the file for saving; this will throw any errors before we
        # do any real work.
        if filename.endswith(".gz"):
            try:
                import gzip
                f = gzip.open(filename, 'wb')
            except ImportError:
                raise SystemExit("Saving gzipped file '%s' requires module gzip" % filename)
        else:
            f = open(filename, 'wb')

        self.build_ovtable()

        # write T, the number of triangles in the model
        f.write("%d\n" % len(self.faces))

        # write each V O pair
        for vertices, opposites in zip(self.faces, self.ovtable):
            for v, o in zip(vertices, opposites):
                f.write("%d %d\n" % (v, o))

        # write V, the number of vertices
        f.write("%d\n" % len(self.vertices))
        for v in self.vertices:
            f.write("%f %f %f\n" % tuple(v))

        f.close()

    def load_ovtable(self, filename):
        """load_ovtable(self, filename) -> None

        Loads OVTable format model from file, replacing current model"""

        if filename.endswith('.gz'):
            try:
                import gzip
                f = gzip.open(filename)
            except ImportError:
                raise SystemExit("Loading gzipped file '%s' requires module gzip" % filename)
        else:
            f = open(filename)
        linecount = 0
        t = int(f.readline())

        self.vertices = []
        self.faces = []
        self.ovtable = []

        for i in range(t):
            self.faces.append([])
            self.ovtable.append([])
            for index in range(3):
                v, o = [int(n) for n in f.readline().split()]
                self.faces[-1].append(v)
                self.ovtable.append(o)

        v = float(f.readline())
        for i in range(v):
            self.vertices.append([float(x) for x in f.readline().split()])

        f.close()
        for attr in "drawlist_face drawlist_texture drawlist_wire drawlist_dual drawlist_normals".split():
            if hasattr(self, attr):
                delattr(self, attr)

        self.calculate_normals()

    def build_vtof(self):
        """build_vtof(self) -> array of arrays of face indices"""

        sf = self.faces
        vtof =[ [] for i in range(len(self.vertices)) ]
        for f in range(len(sf)):
            vtof[sf[f][0]].append(f)
            vtof[sf[f][1]].append(f)
            vtof[sf[f][2]].append(f)

        return vtof

    def build_dual_graph(self):
        """build_dual_graph(self) -> None

        Build the dual graph of the face edge graph."""

        self.dual_face_cycles = []
        # build a quick access vertex->face list
        sf = self.faces
        vtof = self.build_vtof()

        for v in range(len(vtof)):
            cycles = []
            # order the faces and draw the loop around vertex v
            faces = vtof[v][:]
            vlist = [vnum for f in faces for vnum in sf[f]]
            # generate a count of each vertex used and order by it.
            # This way if there are any boundaries, that vertex will be
            # (1, v) and sorted to the front.  If all (2,x), no prob.
            vcount = [(vlist.count(vnum), vnum) for vnum in vlist]
            vcount.sort()
            v0 = v1 = vcount[0][1]

            visited = [v]
            # get the starting face
            ofaces = [f for f in faces if sf[f].count(v0)][0:1]
            # grab the rest in a loop
            while len(ofaces) < len(faces):
                ofaces.extend([f for f in faces if sf[f].count(v1) and f not in ofaces])
                visited.append(v1)
                try:
                    v1 = [vx for vx in sf[ofaces[-1]] if vx not in visited][0]
                except IndexError: # ugh - code dups. but easier this way
                    # non manifold - we've reached the end of one...
                    cycles.append(tuple(ofaces))

                    # reset for next loop component:
                    faces = [f for f in faces if f not in ofaces]
                    vlist = [vnum for f in faces for vnum in sf[f]]
                    vcount = [(vlist.count(vnum), vnum) for vnum in vlist]
                    vcount.sort()
                    try:
                        v0 = v1 = vcount[0][1]
                    except IndexError:
                        print
                        print 'vtof', vtof
                        print 'vlist', vlist
                        print 'vcount', vcount
                        raise
                    visited = visited[0:1]
                    ofaces = [f for f in faces if sf[f].count(v0)][0:1]

            cycles.append(tuple(ofaces))
            self.dual_face_cycles.append(tuple(cycles))

    def draw(self, *args, **kvargs):
        """Draw the model to the waiting screen"""

        # draw faces or textured faces, wires, normals, duals
        if kvargs.has_key('faces') and kvargs['faces']:
            self._draw_faces(*args, **kvargs)
        elif kvargs.has_key('textures') and kvargs['textures']:
            self._draw_textures(*args, **kvargs)
        if kvargs.has_key('wires') and kvargs['wires']:
            self._draw_wires(*args, **kvargs)
        if kvargs.has_key('normals') and kvargs['normals']:
            self._draw_normals(*args, **kvargs)
        if kvargs.has_key('duals') and kvargs['duals']:
            self._draw_duals(*args, **kvargs)

    def successor(self, face, vertex):
        """successor(self, face, vertex) -> face

        Return the next face around vertex"""

        cycles = self.dual_face_cycles[vertex]
        for cycle in cycles:
            if cycle.count(face):
                successor = cycle[(cycle.index(face)+1) % len(cycle)]
        return successor

    def predecessor(self, face, vertex):
        """predecessor(self, face, vertex) -> face

        Return the preceding face around vertex"""

        cycles = self.dual_face_cycles[vertex]
        for cycle in cycles:
            if cycle.count(face):
                predecessor = cycle[cycle.index(face)-1]
        return predecessor


    def _draw_faces(self, *args, **kvargs):
        """Draw the faces of the model to the screen"""
        try:
            list = self.drawlist_face
        except AttributeError:
            # build a display list for speed; python is too slow
            list = self.drawlist_face = glGenLists(1)

            glNewList(self.drawlist_face, GL_COMPILE)
            # draw it the "hard" way
            glBegin(GL_TRIANGLES)
            for f in range(len(self.faces)):
                glNormal3fv(self.normals[f])
                if len(self.colors):
                    for vert in self.faces[f]:
                        glColor3fv(self.colors[vert])
                        glVertex3fv(self.vertices[vert])
                else:
                    for vert in self.faces[f]:
                        glVertex3fv(self.vertices[vert])
            glEnd()
            glEndList()

        glPushAttrib(GL_POLYGON_BIT)
        glPolygonMode(GL_FRONT, GL_FILL)
        glPolygonMode(GL_BACK, GL_LINE)
        #glPolygonOffset(1.0,1.0)
        glLineWidth(1)
        glCallList(list)
        glPopAttrib()

    def _draw_textures(self, *args, **kvargs):
        """Draw the textured faces of the model to the screen"""
        try:
            list = self.drawlist_texture
        except AttributeError:
            # build a display list for speed; python is too slow
            list = self.drawlist_texture = glGenLists(1)

            glNewList(self.drawlist_texture, GL_COMPILE)
            # draw it the "hard" way
            glPushAttrib(GL_ENABLE_BIT)     # save old enables
            glColor4f(1,1,1,1)
            glEnable(GL_TEXTURE_2D)
            glBindTexture(GL_TEXTURE_2D, self.textureid)
            glBegin(GL_TRIANGLES)
            for f in range(len(self.faces)):
                glNormal3fv(self.normals[f])
                for vert in self.faces[f]:
                    glTexCoord2fv(self.texturecoords[vert])
                    glVertex3fv(self.vertices[vert])
            glEnd()
            glPopAttrib()
            glEndList()

        glCallList(list)

    def _draw_wires(self, *args, **kvargs):
        """Draw the wireframe of the model to the screen"""

        try:
            list = self.drawlist_wire
        except AttributeError:
            # build a display list for speed; python is too slow
            list = self.drawlist_wire = glGenLists(1)

            glNewList(self.drawlist_wire, GL_COMPILE)

            # draw it the "hard" way
            glPushAttrib(GL_ENABLE_BIT|GL_CURRENT_BIT|GL_POLYGON_BIT|GL_LINE_BIT|GL_DEPTH_BUFFER_BIT)
            glEnable(GL_POLYGON_OFFSET_FILL)
            glDisable(GL_LIGHTING)
            #glDisable(GL_DEPTH_TEST)
            glPolygonMode(GL_FRONT, GL_LINE)
            glPolygonMode(GL_BACK, GL_LINE)
            #glPolygonOffset(1.0,1.0)
            glLineWidth(1)

            glBegin(GL_TRIANGLES)
            for f in range(len(self.faces)):
                glNormal3fv(self.normals[f])
                for vert in self.faces[f]:
                    glVertex3fv(self.vertices[vert])
            glEnd()

            glPopAttrib()

            glEndList()

            self._draw_wires(*args, **kvargs)

        background = (kvargs.has_key('faces') and kvargs['faces']) or (kvargs.has_key('textures') and kvargs['textures'])

        color = 0.0, 0.0, 0.0, 0.0
        if not background:
            color = 1.0, 1.0, 1.0, 1.0
        glPushAttrib(GL_CURRENT_BIT)
        glColor4fv(color)
        glCallList(list)
        glPopAttrib()

    def _draw_normals(self, *args, **kvargs):
        """Draw the normals of the model to the screen"""

        try:
            list = self.drawlist_normals
        except AttributeError:
            # build a display list for speed; python is too slow
            list = self.drawlist_normals = glGenLists(1)

            glNewList(self.drawlist_normals, GL_COMPILE)

            # draw it the "hard" way
            glPushAttrib(GL_ENABLE_BIT|GL_CURRENT_BIT|GL_POLYGON_BIT|GL_LINE_BIT|GL_DEPTH_BUFFER_BIT)
            glEnable(GL_POLYGON_OFFSET_FILL)
            glDisable(GL_LIGHTING)
            #glDisable(GL_DEPTH_TEST)
            glPolygonMode(GL_FRONT, GL_LINE)
            glPolygonMode(GL_BACK, GL_POINT)
            #glPolygonOffset(1.0,1.0)
            glLineWidth(1)

            glBegin(GL_LINES)
            for f in range(len(self.faces)):
                glNormal3fv(self.normals[f])
                # draw at centroid
                vertices = [self.vertices[v] for v in self.faces[f]]
                centroid = [(a+b+c)/3 for a,b,c in zip(vertices[0], vertices[1], vertices[2])]
                cennorm = [a+b*0.05 for a,b in zip(centroid, self.normals[f])]

                glVertex3fv(centroid)
                glVertex3fv(cennorm)
            glEnd()

            glPopAttrib()

            glEndList()

        color = 1.0, 0.2, 0.2, 1.0
        glPushAttrib(GL_CURRENT_BIT)
        glColor4fv(color)
        glCallList(list)
        glPopAttrib()

    def _draw_duals(self, *args, **kvargs):
        """Draw the dual graph of the model to the screen"""

        try:
            list = self.drawlist_dual
        except AttributeError:
            # build a display list for speed; python is too slow
            list = self.drawlist_dual = glGenLists(1)

            glNewList(self.drawlist_dual, GL_COMPILE)
            # draw it the "hard" way
            glPushAttrib(GL_ENABLE_BIT|GL_CURRENT_BIT|GL_POLYGON_BIT|GL_LINE_BIT)
            glEnable(GL_POLYGON_OFFSET_FILL)
            glDisable(GL_LIGHTING)
            glPolygonMode(GL_FRONT, GL_LINE)
            glPolygonMode(GL_BACK, GL_LINE)
            #glPolygonOffset(1.0,1.0)
            glLineWidth(1)

            if not len(self.dual_face_cycles): self.build_dual_graph()

            sf = self.faces
            for cycle in self.dual_face_cycles:
                # XXX: what about non-loop case?
                for faces in cycle:
                    glBegin(GL_POLYGON)
                    for face in faces:
                        A = self.vertices[sf[face][0]]
                        B = self.vertices[sf[face][1]]
                        C = self.vertices[sf[face][2]]

                        centroid = [(a+b+c)/3 for a,b,c in zip(A,B,C)]
                        glNormal3fv(self.normals[face])
                        glVertex3fv(centroid)
                    glEnd()

            glPopAttrib()
            glEndList()

        color = 0.2, 0.2, 0.7, 1.0
        glPushAttrib(GL_CURRENT_BIT)
        glColor4fv(color)
        glCallList(list)
        glPopAttrib()

    # edgebreaker methods
    edgebreak = edgebreaker.edgebreak
    load_edgebreaker = edgebreaker.load_edgebreaker
    save_edgebreaker = edgebreaker.save_edgebreaker
