#! /usr/bin/env python
#
#    Implements a simple face-vertex representation of a model
#    Copyright (C) 2002  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import model
import math

try:
    import pygame
    from pygame.locals import *
    from OpenGL.GL import *
except:
    print 'msp requires pygame and pyopengl'
    raise SystemExit


def _cross(a, b):
    return (a[1]*b[2]-a[2]*b[1], 
            a[2]*b[0]-a[0]*b[2], 
            a[0]*b[1]-a[1]*b[0])

class Vec:
    def __init__(self, *a):
        self._v = tuple(a[0])

    def dim(self):
        return len(self._v)

    def __mul__(s, o):
        assert len(s) == len(o)
        try:
            len(o)
            dot = 0;
            for x, y in zip(s._v, o._v):
                dot += x*y
            return dot
        except TypeError:
            return Vec([x * o for x in s._v])

    def __rmul__(s, o):
        return Vec([x * o for x in s._v])

    def __sub__(s, o):
        assert len(s) == len(o)
        return Vec([x - y for x, y in zip(s._v, o._v)])

    def __add__(s, o):
        assert len(s) == len(o)
        return Vec([x + y for x, y in zip(s._v, o._v)])

    def __div__(s,d):
        return Vec([x / d for x in s._v])

    def cross(s, o):
        assert len(s) == len(o) == 3
        return Vec(_cross(s._v, o._v))

    def __getitem__(s,i):
        return s.v[i]

    def __len__(s):
        return len(s._v)

    def __abs__(s):
        return math.sqrt(s*s)

    def unit(s):
        return s / len(s)

    def proj(s, o):
        """project s onto o"""
        return (abs(s) * (s * o)) * o


class MSP(model.Model):
    """Handles loading and drawing of a simple model format"""

    def __init__(self, filename=None):
        """MSP([filename]) optionally loads model stored in passed filename"""

        model.Model.__init__(self)

        self.vertices = []          # list of vertices
        self.faces = []             # list of references to vertices
        self.normals = []           # normals correspond to each face
        self.parents = []
        self.parentpos = []
        self.displace = []
        self.msporder = []

        if filename:
            self.load(filename)

    def load(self, filename):
        """load(self, filename) -> None

        Load a model in qslim LOG format"""

        if filename.endswith('.gz'):
            try:
                import gzip
                f = gzip.open(filename)
            except ImportError:
                raise SystemExit("Viewing gzipped files requires module gzip")
        else:
            f = open(filename)
        linecount = 0
        for line in f.readlines():
            linecount += 1
            items = line.split()
            # v X Y Z defines a vertex at (x, y, z)
            if len(items) == 4 and items[0] == 'v':
                self.vertices.append([float(x) for x in items[1:4] ])
                self.parents.append(None)
                self.parentpos.append([])
                self.displace.append(None)
            # f A B C defines a face using vertices A, B, C
            elif len(items) == 4 and items[0] == 'f':
                self.faces.append([int(x)-1 for x in items[1:4] ])
            elif items[0] == 'v%':
                col = [int(x)-1 for x in items[1:3] ]
                pos = [float(x) for x in items[3:6] ]
                self.parents[ col[1] ] = col[0]
                self.parentpos[ col[1] ] = pos
                self.msporder.append(col[1])
            elif line[0] == '#' or len(items) == 0:
                pass
            elif items[0] == 'end' or items[0] == 'begin':
                pass
            else:
                raise RuntimeError("Invalid syntax on line %d '%s'"%(linecount, line))
        f.close()

        self.calculate_normals()
        print 'hello'
        self.calculate_displacements()
        print 'bye'

    def calculate_normals(self):
        """calculate_normals(self) -> None

        Calculates the normals from a proper fvlist"""

        self.normals = []
        for A,B,C in self.faces:
            A, B, C = self.vertices[A], self.vertices[B], self.vertices[C]
            AB = A[0]-B[0], A[1]-B[1], A[2]-B[2]
            AC = A[0]-C[0], A[1]-C[1], A[2]-C[2]
            n = _cross(AB, AC)
            mag = math.sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2])
            try:
                n = n[0]/mag, n[1]/mag, n[2]/mag
            except ZeroDivisionError:
                n = (0.0, 0.0, 0.0)
            self.normals.append(n)

    def calculate_displacements(self):
        """calculate_displacements(self) -> None

        Calculates the frames and displacements of every point with a
        great-grandparent"""

        for vert in self.msporder:
            try:
                v1 = self.parents[vert]
                v2 = self.parents[v1]
                v3 = self.parents[v2]
            except TypeError:
                continue
            if v3 is None: continue

            p0 = Vec(self.vertices[vert])
            p1 = Vec(self.vertices[v1])
            p2 = Vec(self.vertices[v2])
            p3 = Vec(self.vertices[v3])

            x = (p2 - p1).unit()
            y = (p3 - p1).unit()
            z = x.cross(y).unit()
            d = p0 - p1
            dx = d.proj(x)
            dy = d.proj(y)
            dz = d.proj(z)
            self.displace[vert] = (dx, dy, dz)

    def build_vtof(self):
        """build_vtof(self) -> array of arrays of face indices"""

        sf = self.faces
        vtof =[ [] for i in range(len(self.vertices)) ]
        for f in range(len(sf)):
            vtof[sf[f][0]].append(f)
            vtof[sf[f][1]].append(f)
            vtof[sf[f][2]].append(f)

        return vtof

    def build_dual_graph(self):
        """build_dual_graph(self) -> None

        Build the dual graph of the face edge graph."""

        self.dual_face_cycles = []
        # build a quick access vertex->face list
        sf = self.faces
        vtof = self.build_vtof()

        for v in range(len(vtof)):
            cycles = []
            # order the faces and draw the loop around vertex v
            faces = vtof[v][:]
            vlist = [vnum for f in faces for vnum in sf[f]]
            # generate a count of each vertex used and order by it.
            # This way if there are any boundaries, that vertex will be
            # (1, v) and sorted to the front.  If all (2,x), no prob.
            vcount = [(vlist.count(vnum), vnum) for vnum in vlist]
            vcount.sort()
            v0 = v1 = vcount[0][1]

            visited = [v]
            # get the starting face
            ofaces = [f for f in faces if sf[f].count(v0)][0:1]
            # grab the rest in a loop
            while len(ofaces) < len(faces):
                ofaces.extend([f for f in faces if sf[f].count(v1) and f not in ofaces])
                visited.append(v1)
                try:
                    v1 = [vx for vx in sf[ofaces[-1]] if vx not in visited][0]
                except IndexError: # ugh - code dups. but easier this way
                    # non manifold - we've reached the end of one...
                    cycles.append(tuple(ofaces))

                    # reset for next loop component:
                    faces = [f for f in faces if f not in ofaces]
                    vlist = [vnum for f in faces for vnum in sf[f]]
                    vcount = [(vlist.count(vnum), vnum) for vnum in vlist]
                    vcount.sort()
                    try:
                        v0 = v1 = vcount[0][1]
                    except IndexError:
                        print
                        print 'vtof', vtof
                        print 'vlist', vlist
                        print 'vcount', vcount
                        raise
                    visited = visited[0:1]
                    ofaces = [f for f in faces if sf[f].count(v0)][0:1]

            cycles.append(tuple(ofaces))
            self.dual_face_cycles.append(tuple(cycles))

    def draw(self, *args, **kvargs):
        """Draw the model to the waiting screen"""

        # draw faces or textured faces, wires, normals, duals
        if kvargs.has_key('faces') and kvargs['faces']:
            self._draw_faces(*args, **kvargs)
        if kvargs.has_key('wires') and kvargs['wires']:
            self._draw_wires(*args, **kvargs)
        if kvargs.has_key('normals') and kvargs['normals']:
            self._draw_normals(*args, **kvargs)

    def _draw_faces(self, *args, **kvargs):
        """Draw the faces of the model to the screen"""
        try:
            list = self.drawlist_face
        except AttributeError:
            # build a display list for speed; python is too slow
            list = self.drawlist_face = glGenLists(1)

            glNewList(self.drawlist_face, GL_COMPILE)
            # draw it the "hard" way
            glBegin(GL_TRIANGLES)
            for f in range(len(self.faces)):
                glNormal3fv(self.normals[f])
                for vert in self.faces[f]:
                    glVertex3fv(self.vertices[vert])
            glEnd()
            glEndList()

        glPushAttrib(GL_POLYGON_BIT)
        glPolygonMode(GL_FRONT, GL_FILL)
        glPolygonMode(GL_BACK, GL_LINE)
        #glPolygonOffset(1.0,1.0)
        glLineWidth(1)
        glCallList(list)
        glPopAttrib()

    def _draw_wires(self, *args, **kvargs):
        """Draw the wireframe of the model to the screen"""

        try:
            list = self.drawlist_wire
        except AttributeError:
            # build a display list for speed; python is too slow
            list = self.drawlist_wire = glGenLists(1)

            glNewList(self.drawlist_wire, GL_COMPILE)

            # draw it the "hard" way
            glPushAttrib(GL_ENABLE_BIT|GL_CURRENT_BIT|GL_POLYGON_BIT|GL_LINE_BIT|GL_DEPTH_BUFFER_BIT)
            glEnable(GL_POLYGON_OFFSET_FILL)
            glDisable(GL_LIGHTING)
            #glDisable(GL_DEPTH_TEST)
            glPolygonMode(GL_FRONT, GL_LINE)
            glPolygonMode(GL_BACK, GL_LINE)
            #glPolygonOffset(1.0,1.0)
            glLineWidth(1)

            glBegin(GL_TRIANGLES)
            for f in range(len(self.faces)):
                glNormal3fv(self.normals[f])
                for vert in self.faces[f]:
                    glVertex3fv(self.vertices[vert])
            glEnd()

            glPopAttrib()

            glEndList()

            self._draw_wires(*args, **kvargs)

        background = (kvargs.has_key('faces') and kvargs['faces']) or (kvargs.has_key('textures') and kvargs['textures'])

        color = 0.0, 0.0, 0.0, 0.0
        if not background:
            color = 1.0, 1.0, 1.0, 1.0
        glPushAttrib(GL_CURRENT_BIT)
        glColor4fv(color)
        glCallList(list)
        glPopAttrib()

    def _draw_normals(self, *args, **kvargs):
        """Draw the normals of the model to the screen"""

        try:
            list = self.drawlist_normals
        except AttributeError:
            # build a display list for speed; python is too slow
            list = self.drawlist_normals = glGenLists(1)

            glNewList(self.drawlist_normals, GL_COMPILE)

            # draw it the "hard" way
            glPushAttrib(GL_ENABLE_BIT|GL_CURRENT_BIT|GL_POLYGON_BIT|GL_LINE_BIT|GL_DEPTH_BUFFER_BIT)
            glEnable(GL_POLYGON_OFFSET_FILL)
            glDisable(GL_LIGHTING)
            #glDisable(GL_DEPTH_TEST)
            glPolygonMode(GL_FRONT, GL_LINE)
            glPolygonMode(GL_BACK, GL_POINT)
            #glPolygonOffset(1.0,1.0)
            glLineWidth(1)

            glBegin(GL_LINES)
            for f in range(len(self.faces)):
                glNormal3fv(self.normals[f])
                # draw at centroid
                vertices = [self.vertices[v] for v in self.faces[f]]
                centroid = [(a+b+c)/3 for a,b,c in zip(vertices[0], vertices[1], vertices[2])]
                cennorm = [a+b*0.05 for a,b in zip(centroid, self.normals[f])]

                glVertex3fv(centroid)
                glVertex3fv(cennorm)
            glEnd()

            glPopAttrib()

            glEndList()

        color = 1.0, 0.2, 0.2, 1.0
        glPushAttrib(GL_CURRENT_BIT)
        glColor4fv(color)
        glCallList(list)
        glPopAttrib()
