#! /usr/bin/env python
#
#    Implements a transparent 2D overlay drawn as a texture
#    Copyright (C) 2002  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import pygame
import model
import pygame.locals as SDL
import OpenGL.GL
GL = OpenGL.GL

class Overlay(model.Model, pygame.surface.Surface):
    """A 2D image (sprite) that can be drawn to the OpenGL screen"""

    _projection2d = 0

    def __init__(self, *args, **kvargs):
        """Overlay([filename]) optionally loads a "model" image"""

        model.Model.__init__(self, *args, **kvargs)
        pygame.surface.Surface.__init__(self, *args, **kvargs)

    def draw(self, destpos, sourcerect=None):
        _projection2d = Overlay._projection2d
        if not _projection2d:
            GL.glPushAttrib(GL.GL_ENABLE_BIT)     # save old enables
            GL.glDisable(GL.GL_DEPTH_TEST)
            GL.glDisable(GL.GL_CULL_FACE)
            GL.glEnable(GL.GL_TEXTURE_2D)

            GL.glMatrixMode(GL.GL_PROJECTION)
            GL.glPushMatrix()
            GL.glLoadIdentity()
            GL.glOrtho(0.0, 640.0, 480.0, 0.0, 0.0, 1.0)

            GL.glMatrixMode(GL.GL_MODELVIEW)
            GL.glPushMatrix()
            GL.glLoadIdentity()
            Overlay._projection2d = 1

            #glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL)

        if self.get_flags() & SDL.SRCALPHA:
            GL.glColor4f(1,1,1,1)
            GL.glEnable(GL.GL_BLEND)
            GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
        else:
            alpha = self.get_alpha() or 255
            alpha /= 255.0
            GL.glColor4f(1,1,1,alpha)
            GL.glDisable(GL.GL_BLEND)

        x0, y0 = destpos[0:2]
        try:
            x1, y1 = destpos[2:4]
        except IndexError:
            x1 = x0+self.get_width()-1
            y1 = y0+self.get_height()-1

        if sourcerect:
            tx0, ty0, tx1, ty1 = sourcerect
            tx1 = tx1 * float(self.get_width())/self._texture_size[0]
            ty1 = ty1 * float(self.get_height())/self._texture_size[1]

        try:
            self._texture_current
        except AttributeError:
            self._update_texture()

        GL.glBindTexture(GL.GL_TEXTURE_2D, self._texture_id)
        GL.glBegin(GL.GL_TRIANGLE_STRIP)
        GL.glTexCoord2f(tx0, ty0); GL.glVertex2f(x0, y0)
        GL.glTexCoord2f(tx1, ty0); GL.glVertex2f(x1, y0)
        GL.glTexCoord2f(tx0, ty1); GL.glVertex2f(x0, y1)
        GL.glTexCoord2f(tx1, ty1); GL.glVertex2f(x1, y1)
        GL.glEnd()


        if not _projection2d:
            Overlay._projection2d = 0
            GL.glMatrixMode(GL.GL_PROJECTION)
            GL.glPopMatrix()
            GL.glMatrixMode(GL.GL_MODELVIEW)
            GL.glPopMatrix()
            GL.glPopAttrib()

    def _update_texture(self):
        """Update the GL texture so the "blit" draws what we want"""

        w, h = self.get_width(), self.get_height()
        w2, h2 = 1, 1
        while w2 < w: w2 <<= 1
        while h2 < h: h2 <<= 1

        try:
            self._texture_id
        except AttributeError:
            self._texture_id = GL.glGenTextures(1)
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)  
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR) 
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._texture_id

        img = pygame.Surface((w2, h2), GL.SRCALPHA, self)
        img.blit(surf, (0,h2-h), self.get_rect())
        rgba = pygame.image.tostring(img, "RGBA", 0)

        GL.glBindTexture(GL.GL_TEXTURE_2D, self._texture_id)
        if 'glTexSubImage2D' in dir(GL) \
            and w2 <= self._texture_size[0] and h2 <= self._texture_size[1]:

            # untested; i suspect it doesn't work
            w2, h2 = self._texture_size
            glTexSubImage2D(GL_TEXTURE_RECTANGLE_EXT, 0,
                0, 0, w2, h2, GL_RGBA, GL_UNSIGNED_BYTE, rgba);
        else:
            GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA,
                w2, h2, 0, GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, rgba)
            self._texture_size = w2, h2

