import Numeric as N
import math
import types
def _cross(a, b):
    return N.array((\
        a[1]*b[2]-a[2]*b[1], \
        a[2]*b[0]-a[0]*b[2], \
        a[0]*b[1]-a[1]*b[0]))
class Quaternion:
    def __init__(q, quat=None):
        """Quaternion(q, quat=None) -> Quaternion

        Creates a new "pure" quaternion (of no rotation).  If a quaternion
        is passed in, it is copied.  If a tuple of the format (degrees, (x,
        y, z)) or (x, y, z, w) is passed in, it is turned into a
        quaternion."""

        if quat:
            if isinstance(quat, Quaternion):
                q.r = quat.r
                q.v = quat.v[:]
            elif isinstance(quat, types.TupleType):
                if len(quat) == 4:
                    q.v = N.array(quat[0:3])
                    q.r = quat[3]
                elif len(quat) == 2 and isinstance(quat[1], types.TupleType)\
                    and len(quat[1]) == 3:
                        angle = quat[0] * math.pi/360
                        q.r = math.cos(angle)
                        sin_a = math.sin(angle)
                        q.v = sin_a * N.array(quat[1])
                else:
                    raise TypeError("Invalid tuple for argument 2")
            else:
                raise TypeError("Argument 2 must be a tuple")
        else:
            q.v = N.array((1,0,0))
            q.r = 0

    def rotate(self, angle, axis):
        """rotate(self, angle, axis) -> self

        Rotate Quaternion self by angle (degrees) around axis (x, y, z) and
        return self for each chaining"""

        q = Quaternion((angle,axis)) * self
        self.r = q.r
        self.v = q.v

    def __mul__(q, o):
        """Quaternion *= Quaternion -> Quaternion
        Quaternion *= float -> Quaternion

        Multiplies a Quaternion by a constant or by another Quaternion.
        Remember to do (b * a) for a Quaternion representation of a rotation
        by a then by b.  Not Commutative!"""

        if isinstance(o, Quaternion):
            r = q.r; v = q.v
            s = o.r; w = o.v
            angle = r*s - N.dot(v,w)
            vec = r*w + s*v + _cross(v,w)

            n = math.sqrt(angle*angle + N.dot(vec,vec))
            return Quaternion((vec[0]/n, vec[1]/n, vec[2]/n, angle/n))
        else:
            return Quaternion((q.v[0]*o, q.v[1]*o, q.v[2]*o, q.r*o))

    def __imul__(q, o):
        """Quaternion *= Quaternion -> None
        Quaternion *= float -> None

        Multiplies a Quaternion in place by a constant or by another
        Quaternion.  Remember to do (b * a) for a Quaternion representation
        of a rotation by a then by b.  Not Commutative!"""

        if isinstance(o, Quaternion):
            r = q.r; v = q.v
            s = o.r; w = o.v
            q.r = r*s - N.dot(v,w)
            q.v = r*w + s*v + _cross(v,w)
            q.normalize()
        else:
            q.r *= o
            q.v *= o
        return q

    def __abs__(q):
        """abs(Quaternion) -> float

        Returns the magnitue of the Quaternion = sqrt(x*x+y*y+z*z+w*w)"""

        return math.sqrt(q.r*q.r + N.dot(q.v,q.v))

    def normalize(q):
        """normalize(q) -> q

        Normalizes the Quaternion such that the magnitued is 1.0"""

        n = abs(q)
        if n: q *= (1.0/n)
        return q

    def angle(q):
        """angle(q) -> float

        Return the angle of rotation (in degrees) represented by q"""

        return math.acos(q.r)*360/math.pi

    def axis(q):
        """axis(q) -> [x,y,z]

        Returns a (Numeric) array of the axis of rotation represented by q.
        Normalizes the vector to have magnitude of 1.0"""

        n = math.sqrt(N.dot(q.v,q.v))
        if not n: n = 1.0
        return q.v / n

    def __repr__(q):
        """repr(Quaternion) -> string
        
        Return a string of the format '<w [x y z]>' (direct Quaternion
        values)"""

        return '<%f %s>'%(q.r, q.v)

    def __str__(q):
        """str(Quaternion) -> string

        Return a string of the format '<angle (x y z)>' (angle in degrees
        and normalized axis of rotation)"""

        ax = q.axis()
        return '<%0.2f (%0.2f %0.2f %0.2f)>'%(q.angle(), ax[0], ax[1], ax[2])

