#! /usr/bin/env python
#
#    Simple easily extended user interface for viewing models
#    Copyright (C) 2002  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""Handle user interface for displaying models in an easily extended manner"""

import sys, math
try:
    import pygame
    import pygame.locals as SDL
    import OpenGL.GL as GL
except:
    print 'model requires pygame and pyopengl'
    raise SystemExit

try:
    import OpenGL.GLU as GLU
except:
    print "Warning: OpenGL.GLU did not import correctly."
    GLU = None

import sprite

class Handler:
    """Define event handlers by matching all mask's key:val => call callback"""

    def __init__(self, mask, callback):
        """Handler(self, mask, callback) -> Handler

        Create a Handler with mask and callback.  mask should be a dict, and
        callback a function with an event argument"""

        self.mask = mask
        self.callback = callback

class UI:
    """Base User Interface"""

    def add_model(self, model, name):
        """add_model(self, model, name=None) -> None

        Add a model to the scene by name.  Raises KeyError on repeated name."""

        if self.models.has_key(name):
            raise KeyError("Model %s already in use" % name)
        else:
            self.models[name] = model

        if not self.active_model:
            self.active_model = model

    def del_model(self, name):
        """del_model(self, name) -> None

        Delete a model from the scene by name.  Raises KeyError on nonexistent
        name."""

        model = self.models[name]
        del self.models[name]
        if model == self.active_model:
            models = self.models.values()
            if len(models):
                self.active_model = models[0]
            else:
                self.active_model = None

    def scale_active_model(self, S):
        """scale_active_model(self, (x,y,z)) -> None

        Scale the active model by 3-tuple (x,y,z)"""

        self.active_model.scale(S)

    def rotate_active_model(self, theta, axis):
        """rotate_active_model(self, theta, (x,y,z)) -> None

        Rotate the active model by theta around the 3-tuple (x,y,z)"""

        self.active_model.rotate(theta, axis)

    def translate_active_model(self, T):
        """translate_active_model(self, (x,y,z)) -> None

        Translate the active model by 3-tuple (x,y,z)"""

        self.active_model.translate(T)

    def show_caption(self, show):
        """show_caption(self, show) -> None

        Set caption to visible or invisible"""

        self.caption_show = not not show

    def add_handler(self, handler):
        self.handlers.append(handler)

    def clear_handlers(self):
        self.handlers = []

    def setup_gl(self, (width, height)):
        """Set up OpenGL to a simple, sane, state"""

        GL.glClearColor(0.1, 0.4, 0.7, 0.0);
        GL.glEnable(GL.GL_DEPTH_TEST)     # use zbuffer
        GL.glEnable(GL.GL_NORMALIZE)      # normalize normals, etc.

        # turn on ambient light
        GL.glEnable(GL.GL_LIGHTING)
        GL.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, (1.0, 1.0, 1.0, 1.0))
        GL.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, (0.2, -0.2, 1.0, 0.0))
        GL.glEnable(GL.GL_LIGHT0)
        GL.glLightfv(GL.GL_LIGHT1, GL.GL_POSITION, (-0.2, 0.2, 1.0, 0.0))
        GL.glEnable(GL.GL_LIGHT1)

        # set material to a goldish tone
        rgb = (0.912, 0.717, 0.505, 1.0)
        r_amb = tuple([0.1*x for x in rgb])
        r_diff = tuple([1.0*x for x in rgb])
        r_spec = tuple([0.3*x for x in rgb])

        GL.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT, r_amb)
        GL.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_DIFFUSE, r_diff)
        GL.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_SPECULAR, r_spec)
        GL.glMateriali(GL.GL_FRONT_AND_BACK, GL.GL_SHININESS, 100)

        self.reshape((width, height))

    def reshape(self, (width, height)):
        """set up the viewport and projection for width,height"""

        GL.glViewport(0,0, width, height)
        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()

        # set up a simple view: pulled back a bit
        if GLU:
            GLU.gluPerspective(45.0, float(width)/height, 0.1, 100.0)
        else:
            f = 1.3 / (math.tan(45.0/2))
            GL.glMultMatrix((f*height/width, 0, 0, 0,
                        0, f, 0, 0,
                        0, 0, 100.1/-99.9, -1,
                        0, 0, 2*100*0.1/-99.9, 0))
        
        GL.glTranslatef(0.0, 0.0, -3.0)

        # this made mouse rotation harder to understand, so it's gone now.
        #GL.glRotatef(25, 1,0,0) # rotate up a bit

        # return to MODELVIEW so that future transforms are on models
        GL.glMatrixMode(GL.GL_MODELVIEW)

    def cb_update(self, event):
        """set update flag"""
        self.should_update = 1

    def cb_quit(self, event):
        """set to quit"""
        self.should_quit = 1

    def cb_resize(self, event):
        self.reshape(event.size)
        self.cb_update(None)

    def cb_toggle_fullscreen(self, event):
        raise NotImplementedError("cb_toggle_fullscreen is UI-dependent")

    def cb_toggle_wireframe(self, event):
        """toggle active model's draw_wires attribute"""
        self.active_model.draw_wires = not self.active_model.draw_wires
        self.cb_update(None)

    def cb_toggle_faces(self, event):
        """toggle active model's draw_faces attribute"""
        self.active_model.draw_faces = not self.active_model.draw_faces
        self.cb_update(None)

    def cb_toggle_duals(self, event):
        """toggle active model's draw_duals attribute"""
        self.active_model.draw_duals = not self.active_model.draw_duals
        self.cb_update(None)

    def cb_toggle_normals(self, event):
        """toggle active model's draw_normals attribute"""
        self.active_model.draw_normals = not self.active_model.draw_normals
        self.cb_update(None)

    def cb_toggle_textures(self, event):
        """toggle active model's draw_faces attribute"""
        self.active_model.draw_textures = not self.active_model.draw_textures
        self.cb_update(None)

    def cb_select_next_model(self, event):
        models = self.models.values()

        am = self.active_model
        self.active_model = models[(models.index(am)+1) % len(models)]

    def cb_default_mouse(self, event):
        raise NotImplementedError("cb_default_mouse is UI-dependent")

    def _cb_apply_mouse_motion(self):
        dx, dy = self.motion
        self.motion = [0,0]

        if self.motion_buttons[0]: # rotate
            dist = math.sqrt(dx*dx+dy*dy)
            self.rotate_active_model(dist, (dy, dx, 0))
        elif self.motion_buttons[1]: # scale
            s = math.exp((dx+dy)*0.01)
            self.scale_active_model((s,s,s))
        elif self.motion_buttons[2]: # translate
            self.translate_active_model((dx*0.02, -dy*0.02, 0))


class pygameUI(UI):
    """PYGAME User Interface"""

    def run(self, **kvargs):
        """run(self, **kvargs) -> None

        Initializes the display and runs until finished.
        
        Optional keywords include size:(x,y), and fullscreen:bool"""

        winsize = kvargs.has_key('size') and kvargs['size'] or (640, 480)
        fullscreen = kvargs.has_key('fullscreen') and kvargs['fullscreen'] and SDL.FULLSCREEN or 0

        self.screenmode = fullscreen|SDL.OPENGL|SDL.DOUBLEBUF|SDL.RESIZABLE
        pygame.display.set_mode(winsize, self.screenmode)

        self.setup_gl(winsize)

        while 1:
            events = [pygame.event.wait()]
            if pygame.event.peek():
                events.extend(pygame.event.get())

            # process all events
            for event in events:
                for handler in self.handlers:
                        for k,v in handler.mask.items():
                            try:
                                if getattr(event, k) != v:
                                    break
                            except AttributeError:
                                break
                        else:
                            if handler.callback(event):
                                break

            # quit?
            if self.should_quit:
                break

            # draw?
            if self.should_update:
                self.should_update = 0
                self._cb_apply_mouse_motion()   # catch up to mouse

                GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
                for model in self.models.values():
                    model.draw()

                if self.caption_show:
                    screen = pygame.display.get_surface()
                    cs = self.caption_sprite
                    cs.enter()
                    cs.blit_at(
                        (screen.get_width()-cs.get_width()+1,
                        screen.get_height()-cs.get_height()+1))
                    cs.exit()

                pygame.display.flip()


    def set_caption(self, caption, color=(1,1,1), font=None, size=48):
        """set_caption(self, caption, color=(1,1,1), font=None, size=48) -> None
        
        Set window caption to caption, with color (default=white)"""

        self.caption_text = caption
        self.caption_color = color
        usefont = pygame.font.Font(font, size)
        self.caption_image = usefont.render(caption, 1, color)
        self.caption_sprite = sprite.sprite(self.caption_image)

    def cb_select_next_model(self, event):
        """select the next model, or previous if shift is being held"""

        delta = 1
        if pygame.key.get_mods() & SDL.KMOD_SHIFT:
            delta = -1
        models = self.models.values()

        am = self.active_model
        self.active_model = models[(models.index(am)+delta) % len(models)]

    def cb_toggle_fullscreen(self, event):
        """toggle fullscreen (only on ALT-enter)"""
        if pygame.key.get_mods() & SDL.KMOD_ALT:
            pygame.display.toggle_fullscreen()
            self.cb_update(None)

    def cb_resize(self, event):
        """resize the window
        
        This requires more steps than you might expect:

        An event of VIDEORESIZE only indicates a request.  To actually get
        pygame/sdl to know the window has changed, it requires a new set_mode().
            
        Then finally a glViewport() call makes GL use the entire area.
        """

        #super(pygameUI, self).cb_resize(event)
        UI.cb_resize(self, event)
        # odd - moving it down here helps it keep up, but doesn't fix it.
        pygame.display.set_mode(event.size, self.screenmode)

    def cb_defaultmouse(self, event):
        """default mouse handling"""
        if event.type == SDL.MOUSEBUTTONDOWN:
            if event.button in (1,2,3):
                self._cb_apply_mouse_motion()
                if not self.mouse_hidden:
                    pygame.mouse.set_visible(0)
                    pygame.event.set_grab(1)
                self.mouse_hidden += 1
            elif event.button in (4,5):
                if event.button == 4:
                    self.translate_active_model((0, 0, 0.1))
                else:
                    self.translate_active_model((0, 0, -0.1))
                self.cb_update(None)
        elif event.type == SDL.MOUSEBUTTONUP:
            if event.button in (1,2,3):
                self._cb_apply_mouse_motion()
                self.mouse_hidden -= 1
                if not self.mouse_hidden:
                    pygame.mouse.set_visible(1)
                    pygame.event.set_grab(0)
        elif event.type == SDL.MOUSEMOTION:
            if self.mouse_hidden:
                self.motion_buttons = event.buttons
                self.motion[0] += event.rel[0]
                self.motion[1] += event.rel[1]
                self.cb_update(None)

    def __init__(self):
        """__init__(self) -> pygameUI

        Initialize a pygameUI with default settings, handlers"""

        pygame.display.init()
        pygame.font.init()

        # default values
        self.set_caption('model.UI')
        self.caption_show = 0       # don't show some crazy caption
        self.models = {}            # no models to start with
        self.should_quit = 0        # don't quit until requested
        self.should_update = 1      # start with an update to draw scene
        self.mouse_hidden = 0       # mouse should normally be visible
        self.motion = [0,0]         # accumulated motion starts at 0
        self.motion_buttons=[0,0,0] # mouse buttons start unpressed
        self.active_model = None    # no models means no active model

        # default handlers
        self.default_handlers = [
            # quit on close or ESC
            Handler({'type':SDL.QUIT}, self.cb_quit),
            Handler({'type':SDL.KEYDOWN, 'key':SDL.K_ESCAPE}, self.cb_quit),

            # toggle fullscreen on RETURN
            Handler({'type':SDL.KEYDOWN, 'key':SDL.K_RETURN},
                    self.cb_toggle_fullscreen),

            # toggle wireframe on W, faces on F, textures on T
            # duals on D, normals on N
            Handler({'type':SDL.KEYDOWN, 'key':SDL.K_w},
                    self.cb_toggle_wireframe),
            Handler({'type':SDL.KEYDOWN, 'key':SDL.K_f},
                    self.cb_toggle_faces),
            Handler({'type':SDL.KEYDOWN, 'key':SDL.K_t},
                    self.cb_toggle_textures),
            Handler({'type':SDL.KEYDOWN, 'key':SDL.K_d},
                    self.cb_toggle_duals),
            Handler({'type':SDL.KEYDOWN, 'key':SDL.K_n},
                    self.cb_toggle_normals),

            # next model selected on TAB
            Handler({'type':SDL.KEYDOWN, 'key':SDL.K_TAB},
                    self.cb_select_next_model),

            # redraw when exposed, resized
            Handler({'type':SDL.VIDEOEXPOSE}, self.cb_update),
            Handler({'type':SDL.VIDEORESIZE}, self.cb_resize),

            # mouse handlers
            Handler({'type':SDL.MOUSEBUTTONDOWN}, self.cb_defaultmouse),
            Handler({'type':SDL.MOUSEBUTTONUP}, self.cb_defaultmouse),
            Handler({'type':SDL.MOUSEMOTION}, self.cb_defaultmouse),
        ]

        self.handlers = self.default_handlers[:]
